 
#include "i2c.h"

#define XFER_COMPLETE                  ( I2C_IBSR_IBIF_MASK | I2C_IBSR_TCF_MASK )
#define READ_BIT                       0x01


// returns ack status (uses same i2c.h return codes)
static int32_t  wait_for_xfer_complete (
  I2C_MemMapPtr     module )
{
  uint8_t                    status ;
  
  // wait for transfer to complete (IBIF and TCF bits are set)
  // it never hangs, always generates ack clock and samples ack time slot
  do
  {
    status = module->IBSR ;
  } while ( XFER_COMPLETE != ( XFER_COMPLETE & status ) ) ;
  
  module->IBSR |= I2C_IBSR_IBIF_MASK ; // weird: clear IBIF bit by writing a 1 to it
  
  if ( I2C_IBSR_RXAK_MASK & status ) // did slave ack?
    return ( I2C_NO_ACK ) ;
  else
    return ( I2C_SUCCESS ) ;
}

static void  issue_stop (
  I2C_MemMapPtr     module )
{
  uint8_t                    status ;
  
  module->IBCR &= ~I2C_IBCR_MSSL_MASK ; // exit master mode to cause a stop
  
  // IBB bit is set when a stop condition is detected.  we need this to handle the issue
  // of back to back stop/start.  if a start is issued before the stop can clock out, bus hangs
  do
  {
    status = module->IBSR ;
  } while ( I2C_IBSR_IBB_MASK & status ) ;  
}


// Initializes the I2C in master mode
extern void i2c_init ( I2C_MemMapPtr module )
{
  // USING I2C0
  
  reg32_write ( IOMUXC_I2C0_IPP_SCL_IND_SELECT_INPUT , 0x01 ) ; // Alt2 so PTB14 = SCL  
  reg32_write ( IOMUXC_I2C0_IPP_SDA_IND_SELECT_INPUT , 0x01 ) ; // Alt2 so PTB15 = SDA

  reg32_write ( IOMUXC_PTB14 ,
    ( 2 << IOMUXC_PTn_MUX_MODE_SHIFT ) |   // Mux Mode = SCL of I2C0
    ( 0 << IOMUXC_PTn_SPEED_SHIFT    ) |   // Speed = Low, 50 MHz
    ( 0 << IOMUXC_PTn_SRE_SHIFT      ) |   // Slew Rate = slow
    ( 1 << IOMUXC_PTn_ODE_SHIFT      ) |   // Open Drain Enable = open drain
    ( 0 << IOMUXC_PTn_HYS_SHIFT      ) |   // Hysteresis = CMOS input
    ( 4 << IOMUXC_PTn_DSE_SHIFT      ) |   // Drive Strength = 37 ohm (of 20-150 ohms)
    ( 0 << IOMUXC_PTn_PUS_SHIFT      ) |   // Pull Up/Down = 100K pull down
    ( 0 << IOMUXC_PTn_PKE_SHIFT      ) |   // Pull/Keep Enable = disabled
    ( 0 << IOMUXC_PTn_PUE_SHIFT      ) |   // Pull/Keep Select = keeper
    ( 1 << IOMUXC_PTn_OBE_SHIFT      ) |   // Output Buffer Enable = enabled
    ( 1 << IOMUXC_PTn_IBE_SHIFT      ) ) ; // Input Buffer Enable = enabled

  reg32_write ( IOMUXC_PTB15 ,
    ( 2 << IOMUXC_PTn_MUX_MODE_SHIFT ) |   // Mux Mode = SDA of I2C0
    ( 0 << IOMUXC_PTn_SPEED_SHIFT    ) |   // Speed = Low, 50 MHz
    ( 0 << IOMUXC_PTn_SRE_SHIFT      ) |   // Slew Rate = slow
    ( 1 << IOMUXC_PTn_ODE_SHIFT      ) |   // Open Drain Enable = open drain
    ( 0 << IOMUXC_PTn_HYS_SHIFT      ) |   // Hysteresis = CMOS input
    ( 4 << IOMUXC_PTn_DSE_SHIFT      ) |   // Drive Strength = 37 ohm (of 20-150 ohms)
    ( 0 << IOMUXC_PTn_PUS_SHIFT      ) |   // Pull Up/Down = 100K pull down
    ( 0 << IOMUXC_PTn_PKE_SHIFT      ) |   // Pull/Keep Enable = disabled
    ( 0 << IOMUXC_PTn_PUE_SHIFT      ) |   // Pull/Keep Select = keeper
    ( 1 << IOMUXC_PTn_OBE_SHIFT      ) |   // Output Buffer Enable = enabled
    ( 1 << IOMUXC_PTn_IBE_SHIFT      ) ) ; // Input Buffer Enable = enabled
  
  /* Enable the clock for the I2C module */
  if (module == I2C0)
    CCM->CCGR4 |= CCM_CCGR4_CG70_MASK;
  if (module == I2C1)
    CCM->CCGR4 |= CCM_CCGR4_CG71_MASK;
  if (module == I2C2)
    CCM->CCGR4 |= CCM_CCGR4_CG72_MASK;
  if (module == I2C3)
    CCM->CCGR4 |= CCM_CCGR4_CG73_MASK;  
    
  // Set the divider for the SCL frequency
  // 0x38 --> 64.9 KHz
  // 0x24 --> 280 KHz?
  module->IBFD = 0x38 ;
    
  // Set the slave address - won't actually be used
  module->IBAD = 0xC0 ;
    
  // Clear MDIS to enable I2C
  module->IBCR &= ~I2C_IBCR_MDIS_MASK ;
    
  // enable interrupt generation (weird: required for IBIF to be set)
  module->IBCR |= I2C_IBCR_IBIE_MASK ;
    
  // set slave mode so we don't generate a start condition accidentally
  module->IBCR &= ~I2C_IBCR_MSSL_MASK ;

  // disable DMA
  module->IBCR &= ~I2C_IBCR_DMAEN_MASK ;
}  
 

extern int32_t  i2c_write (
  I2C_MemMapPtr     module ,
  uint8_t           slave_addr ,
  uint8_t           reg_num ,
  uint8_t *         buf ,
  uint16_t          buf_len )
{
  uint16_t                   i ;
  uint32_t                   rtn ;
  
  module->IBCR |= I2C_IBCR_MSSL_MASK ; // enter master mode to cause a start
  
  module->IBCR |= I2C_IBCR_TXRX_MASK ; // set transmit mode

  // send slave address, indicate write:
  module->IBSR |= I2C_IBSR_IBIF_MASK ; // weird: clear IBIF bit by writing a 1 to it
  module->IBDR = slave_addr ;
  rtn = wait_for_xfer_complete ( module ) ;
  if ( rtn == I2C_NO_ACK ) // did slave ack?
    return ( rtn ) ;
    
  // send reg number:
  module->IBSR |= I2C_IBSR_IBIF_MASK ; // weird: clear IBIF bit by writing a 1 to it
  module->IBDR = reg_num ;
  rtn = wait_for_xfer_complete ( module ) ;
  if ( rtn == I2C_NO_ACK ) // did slave ack?
    return ( rtn ) ;  
  
  for ( i = 0 ; i < buf_len ; i++ )
  {
    // send one data byte:
    module->IBSR |= I2C_IBSR_IBIF_MASK ; // weird: clear IBIF bit by writing a 1 to it
    module->IBDR = buf[i] ;
    rtn = wait_for_xfer_complete ( module ) ;
    if ( rtn == I2C_NO_ACK ) // did slave ack?
      return ( rtn ) ;
  }
  issue_stop ( module ) ;
  
  return ( I2C_SUCCESS ) ;
}  

extern int32_t  i2c_read (
  I2C_MemMapPtr     module ,
  uint8_t           slave_addr ,
  uint8_t           reg_num ,
  uint8_t *         buf ,
  uint16_t          buf_len )
{  
  uint16_t                   i ;
  int32_t                    rtn ;
  
  module->IBCR |= I2C_IBCR_MSSL_MASK ; // enter master mode to cause a start
  
  module->IBCR |= I2C_IBCR_TXRX_MASK ; // set transmit mode

  // send slave address, indicate write:
  module->IBSR |= I2C_IBSR_IBIF_MASK ; // weird: clear IBIF bit by writing a 1 to it
  module->IBDR = slave_addr ;
  rtn = wait_for_xfer_complete ( module ) ;
  if ( rtn == I2C_NO_ACK ) // did slave ack?
    return ( rtn ) ;   

  // send reg number:
  module->IBSR |= I2C_IBSR_IBIF_MASK ; // weird: clear IBIF bit by writing a 1 to it
  module->IBDR = reg_num ;
  rtn = wait_for_xfer_complete ( module ) ;
  if ( rtn == I2C_NO_ACK ) // did slave ack?
    return ( rtn ) ;  
  
  module->IBCR |= I2C_IBCR_RSTA_MASK ; // issue repeated start
  
  // send slave address, indicate read:
  module->IBSR |= I2C_IBSR_IBIF_MASK ; // weird: clear IBIF bit by writing a 1 to it
  module->IBDR = slave_addr | READ_BIT ;
  rtn = wait_for_xfer_complete ( module ) ;
  if ( rtn == I2C_NO_ACK ) // did slave ack?
    return ( rtn ) ;   
  
  module->IBCR &= ~I2C_IBCR_TXRX_MASK ; // set receive mode

  // receive data one byte at a time.
  for ( i = 0 ; i < buf_len ; i++ )
  {
    if ( i == ( buf_len - 1 ) ) // don't send an ack on last byte
      module->IBCR |= I2C_IBCR_NOACK_MASK ;
    else
      module->IBCR &= ~I2C_IBCR_NOACK_MASK ;
 
    if ( i == 0 )
      buf[0] = module->IBDR ; // dummy read just to start receiving
    else
      buf[i-1] = module->IBDR ;
    
    rtn = wait_for_xfer_complete ( module) ;
  }
  
  // end bus cycle so the following read does not cause another bus cycle
  issue_stop ( module ) ;
  buf[i-1] = module->IBDR ; // must get the results of the last read cycle
     
  return ( I2C_SUCCESS ) ;
}